/*
** Module   : braces.g
** Abstract : This grammar parses the Progress Preprocessor {...} language
**            elements and replaces them with evaluations on the input
**            stream. Unlike the other grammars, this one works independently
**            from the main line.
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 NVS 20041112   @18834 Created. This grammar has to be separate as
**                           it gets activated and terminated
**                           independently from the main stream text
**                           parser and lexer, from the ClearStream class,
**                           transparently to the rest of the
**                           preprocessor.
** 002 NVS 20050127   @19417 Lexer vocabulary was for 0x03-0xff which was
**                           causing problems converting "~000". Fixed to
**                           be the full range 0x00-0xff.
** 003 NVS 20050131   @19462 As it has turned out, the syntax for named
**                           arguments in the include file specification 
**                           allows white spaces before and AFTER the '='
**                           sign. The existing rule for argument value
**                           didn't match that case. Modified the argvalue
**                           rule so it accounts for possible white spaces
**                           and they don't make it into the value.
** 004 NVS 20050216   @19836 Fixed bug in parsing of argument values coded
**                           as strings. The old logic handled cases like
**                           "" (empty string) in a wrong way producing
**                           " as the value. The new logic is: first strip
**                           off the outer quotes, if any; second delete
**                           the extra inner quotes. This matches the
**                           Progress results.
** 005 NVS 20050221   @19902 Backed out the change #004. There was no bug
**                           in the parsing of arguments, just a wrong
**                           interface between the RUN arguments processor
**                           (from hints file) and the FileScope 
**                           constructor.
** 006 NVS 20050513   @21163 Many methods now throw IOException. Added to
**                           method headers and javadoc.
** 007 GES 20070326   @32640 Many format cleanups. Fixes to how arguments
**                           are parsed. The first argument determines
**                           the parsing strategy. If positional, then
**                           all subsequent arguments (even if they have
**                           the &name=value form) will be parsed as
**                           positional (so the &name= text will be part
**                           of the argument). If the first argument is
**                           named (prefixed with &name=), then any stray
**                           text in between properly formatted named
**                           arguments will be parsed as positional
**                           args and discarded as trash.  Added lots of
**                           documentation.
** 008 GES 20070923   @35186 Print a stack trace on errors.
** 009 GES 20070927   @35255 Handle a case of malformed named arguments
**                           where there is no EQUALS following the name.
**                           That argument and all others are then
**                           dropped.
** 010 GES 20070927   @35261 Remove all forms of whitespace from named
**                           argument names, not just spaces.
** 011 GES 20071204   @36191 Added support to disable/enable expansions.
**                           Also reworked the argument value processing
**                           to allow for an empty argument value by
**                           encountering the RBRACE after whitespace.
**                           Modified named arguments parsing to discard
**                           duplicates of named arguments so that the
**                           first instance is the one honored. Major
**                           refactoring to change generated prediction
**                           logic such that a reliable prediction is
**                           always obtained. When the parser exits, it
**                           pushs back any characters that have been
**                           read by the lexer. RBRACE is now matched
**                           inside the CODE rule. These changes enable
**                           recursive includes, includes in between a
**                           '/' and '*' (open comment) and other complex
**                           cases.
** 012 GES 20071205   @36214 Added full support for referencing a
**                           duplicate named argument via its positional
**                           name and via {*}.
** 013 GES 20080318   @37642 Changed approach to read characters (the
**                           entire unicode charset) instead of dealing
**                           with bytes. This makes it safe to process
**                           files with I18N requirements.
** 014 GES 20081010   @40332 Handled additional lookahead pushback case.
**                           Escape character processing flaw fix in the
**                           lexer CODE rule.
** 015 GES 20100225   @44690 Minor comment fix.
** 016 ECF 20101001          Fixed nameref rule. Did not handle case of a
**                           valid nameref, followed by whitespace and then a
**                           closing brace (without additional, non-whitespace
**                           trash in between).
** 017 GES 20110630          Fixed a case of improper parsing for a positional
**                           arg of the form "{&something}" or even just "".
**                           There was logic to remove the double quotes but
**                           that logic would leave any double quote char
**                           behind that was the last char in the arg. The
**                           code was simplified to just never insert the
**                           double quote chars in the first place, since we
**                           were artificially adding them back. I'm not sure
**                           why the code wasn't written that way to begin with.
**                           Refactored the CODE rule in the lexer to match a
**                           single RBRACE char earlier.  This made the code
**                           simpler and reduced the lookahead that was needed
**                           which was causing some problems since it was
**                           looking past the RBRACE which can cause unwanted
**                           (and detrimental) braces parsing recursion.
** 018 GES 20111202          Cleaned up test code and added a comment about a
**                           latent bug.
*/

header {             
/*
** Module   : BracesLexer.java
**            BracesParser.java
**            PreprocTokenTypes.java
** Abstract : This grammar parses the Progress Preprocessor {...} language
**            elements and replaces them with evaluations on the input
**            stream. Unlike the other grammars, this one works independently
**            from the main line.
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20041116 ADD   @18834 WARNING, THIS IS A GENERATED FILE!!!
**                               DO NOT EDIT THIS FILE. The original source
**                               file is braces.g!
*/
package com.goldencode.p2j.preproc;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.util.*;
import antlr.collections.AST;
import antlr.collections.impl.*;
import antlr.debug.misc.*;
import antlr.*;
}

/**
 * Parses all the preprocessor constructs coded in braces which are
 * argument references, preprocessor builtin variable references,
 * defined variable references and the include files.
 * <p>
 * This grammar is not a top level grammar for parsing the Progress
 * source files. The <code>text.g</code> is such a grammar. This grammar
 * is supposed to be used from the <code>{@link ClearStream}</code> class
 * internally.
 * <p>
 * The parsing starts when the driving code calls the {@link #braces} method
 * of this class.
 * <p>
 * No output is written to the output stream during parsing of braces.
 * Everything inside and the braces themselves are dropped from the source
 * and a substitution string is pushed back into the input stream so that
 * the contents of the substitution will be read and interpreted (rescanned)
 * next.
 * <p>
 * As soon as this parser reaches the closing brace token, the parser is
 * done and the calling code resumes execution.
 * <p>
 * The parser calls the lexer's <code>nextToken</code> method and works
 * with the stream of <code>Token</code> objects. So, all the details
 * of tokenizations are hidden in the lexer. The lexer, in turn, reads
 * its input character by character from an input stream. The input
 * stream is an instance of the <code>ClearStream</code> class
 * that hides from the lexer some essential complexity of the preprocessing
 * like handling the alternative codings, escape sequences and substitutions
 * for braces. That automatically means that the nested braces are
 * processed recursively in the <code>ClearStream</code>, so any single
 * instance of the <code>BracesParser</code> has to deal with the non-nested
 * braces exclusively.
 * <p>
 * This parser recognizes and properly handles the following constructs:
 * <ul>
 *   <li>argument references <code>{0}, {n}, {*}, {&*}, {&name}, {}</code>;
 *   <li>defined variable references <code>{&name}</code>;
 *   <li>preprocessor builtin variable references <code>{&name}</code>;
 *   <li>include file references <code>{file}, {file pos}, {file &name=value}
 *       </code> etc;
 * </ul>
 * As the result of an include file reference parsing, no immediate
 * substitute string is produced. Rather, another <code>FileScope</code>
 * instance is created and chained to the new dictionary scope by calling
 * <code>{@link Preprocessor#includeFile}</code> method.
 *
 * @see  ClearStream
 * @see  BracesLexer
 * @see  FileScope
 * @see  TextParser
 */
class BracesParser extends Parser;

options
{
   k = 2;                                          // lookahead depth
   importVocab = Shared;
   exportVocab = Preproc;
   // codeGenBitsetTestThreshold=999;
}

{
   /** Shared state for multiple lexers, parsers and the input stream. */
   private Environment env = null;
   
   /** Token source. */
   private BracesLexer lexer = null;
   
   /** Deferred parser lookahead that must be pushed back. */
   private String parserLookahead = null;

   /**
    * Constructor. Creates a parser attached to the token stream
    * selector taken from the environment. Saves the environment
    * for future needs.
    *
    * @param    env
    *           instance of the preprocessor environment
    * @param    lexer
    *           instance of the lexer to be the source of tokens
    *           for this parser
    */
   public BracesParser(Environment env, BracesLexer lexer)
   {
      this(lexer);
      this.env   = env;
      this.lexer = lexer; 
   }
   
   /**
    * Writes error data to <code>System.err</code>, including a full stack
    * trace.
    *
    * @param    re
    *           The error on which to report.
    */
   public void reportError(RecognitionException re)
   {
      re.printStackTrace(System.err);
   }
}

/**
 * Matches all the preprocessor constructs in braces and substitutes
 * them with their evaluations or activates new input files for the
 * include file references. These curley braces must always come in matching
 * pairs.
 * <p>
 * Valid alternatives include empty braces, braces with only whitespace
 * inside, braces containing an argument reference (named or positional) and
 * braces with an include file reference.  The core processing is handled
 * by {@link #argref} and {@link #incref}.
 * <p>
 * The expansion of braces can be disabled or enabled by using the
 * {@link Environment#setExpandBraces}. Note that no matter the state of the
 * expand flag, the braces and contained text will still be matched. 
 * <p>
 * This is the top level recognizer for all braces.
 */
braces
throws IOException
   {
      // this must be initialized to null to bypass argument processing
      // in the exit action if this is an include or empty braces case
      String str = null;
      
      boolean bypass = false;
   }
   :
      (
         options { generateAmbigWarnings=false; }
         :
         
           // only whitespace
           WS
           
           // argument reference
         | { !(LA(1) == DIGITS && LA(2) != RBRACE) }?
           str = argref                    
           
           // include file reference
         | incref  { bypass = true; }
         
         | // empty braces
      )
      
      RBRACE
      
      {
         // we don't honor isExpandBraces() here because it is possible for
         // an &ENDIF or other preproc directive to be encountered by the
         // expansion of a preproc name or arg reference; so braces expansion
         // flag only disables include file processing (see incref)
         
         StringBuilder sb = new StringBuilder();
         
         // the output from our expansion needs to be pushed back onto the
         // input stream so that when we return from this method (and this
         // instance of the braces parser/lexer pop off the stack), that
         // the braces that we have parsed have been replaced with the
         // proper content (so long as there actually is any content to
         // expand)
         if (str != null)
         {
            // was there any text to expand or was the expansion an empty
            // string?
            if (str.length() > 0)
            {
               sb.append(str);
            }
         }
         
         // include files handle pushback of the lookahead differently to
         // get the ordering of output correct, so bypass this processing
         // in that case, but by default this lookahead pushback is done and
         // this is independent of whether the braces are being expanded or
         // not since in either case the matching may have lookahead that
         // will be lost and which is not part of the braces processing
         if (!bypass)
         {
            // the lexer may have used lookahead in a hard to predict manner
            // which will potentially change with each rebuild, so there may
            // or may not be any characters read from the stream and buffered
            // in the lexer; get any that are there and push them back
            // otherwise they will be lost when we return and the lexer is
            // popped off the stack
            String buffered = env.lookahead(lexer);
            
            if (buffered.length() > 0)
            {
               sb.append(buffered);
            }
         }
      
         // if there is anything to push back, do so
         if (sb.length() > 0)
         {
            env.pushBack(sb.toString());
         }
      }
   ;

/**
 * Matches all the argument references, even misformed ones and returns
 * their evaluations.
 * <p>
 * This is the top level recognizer for all references that produce a
 * substitution (not include files).
 *
 * @return   The argument reference evaluation as a string.
 */
argref returns [String str]
throws IOException
   {
      str = null;
   }
   :
      (
         options { generateAmbigWarnings=false; }
         :
           str = posargref      // positional argument reference  {n}      or
         | str = posallref      // all positional arguments reference {*}  or
         | str = keyallref      // all named arguments reference {&*}      or
         | str = nameref        // named argument reference {&name}        or
         | misformed            // misformed named argument {& ...}        or
      )
   ;

/**
 * Matches positional argument references and returns their evaluations.
 *
 * @return   The positional argument reference evaluation as a string.
 */
posargref returns [String str]
throws IOException
   {
      str = null;
   }
   :
      n:DIGITS               // one or more digits
      {
         str = Preprocessor.getPosArg(env, n.getText());
      }
   ;

/**
 * Matches the all positional argument reference <code>{*}</code>
 * and returns its evaluation.
 *
 * @return   The evaluated <code>{*}</code> as a string.
 */
posallref returns [String str]
   {
      str = null;
   }
   :
      STAR (anything)?  // trash is allowed to follow
      {
         str = Preprocessor.getAllPos(env);
      }
   ;

/**
 * Matches the all named argument reference <code>{&*}</code> and returns
 * its evaluation.
 *
 * @return   The evaluated <code>{&*}</code> as a string.
 */
keyallref returns [String str]
   {
      str = null;
   }
   :
      AMPER STAR (anything)?  // trash is allowed to follow
      {
         str = Preprocessor.getAllNamed(env);
      }
   ;

/**
 * Matches a named argument or variable reference and returns their
 * evaluation.
 *
 * @return   The reference evaluation as a string.
 */
nameref returns [String str]
throws IOException
   {
      str = null;
   }
   :
      AMPER str = anythingButSpace // &name sequence
      (anything)?                  // arbitrary trash is allowed to follow
      {
         str = Preprocessor.getVariable(env, str);
      }
   ;

/**
 * Matches misformed named argument references so they don't break
 * the parsing process and discards them.
 */
misformed
   :
      AMPER WS (anything)?
      {
         env.eprint("misformed variable reference ignored", 0);
      }
   ;

/**
 * Matches any sequence of characters except the closing brace
 * so they don't break the parsing process and discards them.
 */
anything
   :
      (
         options { greedy=true; }
         :
         EQUALS | STAR | AMPER | DIGITS | CODE | QUOTE | WS
      )+
   ;

/**
 * Matches all the include file references and arranges their inclusion
 * into the input stream. The rules {@link #filename} and {@link #arguments}
 * are used to handle the core parsing.
 * <p>
 * This is the top level recognizer for all references that resolve into
 * include files versus substitutions.
 */
incref
   {
      String fname = "";
      Map    args  = new LinkedHashMap();
   }
   :
      fname = filename
      (
         options { generateAmbigWarnings=false; }
         :
           WS
         | arguments[args]
      )?
      {
         // there is a mode where we must parse a given set of braces to
         // know where they end but we do not implement include processing
         // because we are in a skipped conditional block; to do otherwise
         // causes major problems with recursive inclusion (which is allowed
         // in Progress)
         if (env.isExpandBraces())
         {
            String buffered = null;
            
            // the lexer may have used lookahead in a hard to predict manner
            // which will potentially change with each rebuild, so there may
            // or may not be any characters read from the stream and buffered
            // in the lexer; get any that are there and push them back
            // otherwise they will be lost when we return and the lexer is
            // popped off the stack
            
            try
            {
               StringBuffer sb = new StringBuffer();
               
               // we know that LA(1) == RBRACE at this point, but certain
               // use cases of the arguments rule above may have preloaded
               // LA(2); this means that some of the lookahead is no longer
               // in the lexer, but is now in the parser; one case where this
               // happens for sure is when there is a named argument that
               // ends in a QUOTE character (just before the RBRACE)
               if (parserLookahead != null)
               {
                  sb.append(parserLookahead);
                  parserLookahead = null;
               }
               
               // now add any pending lexer lookahead
               sb.append(env.lookahead(lexer));
               
               // remove any markers
               buffered = env.filter(sb, false);
            }
            
            catch (IOException ioe)
            {
               ioe.printStackTrace(System.err);
            }
            
            Preprocessor.includeFile(env, fname, args, buffered);
         }
      }
   ;

/**
 * Matches include file name portion of the include file references
 * and returns the file name.
 *
 * @return   The referenced file name as a string.
 */
filename returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      String txt1 = "";
      String txt2 = "";
      
      str = "";
   }
   :
      (
          options { generateAmbigWarnings=false; }
          :
            // space * anything
            WS STAR txt1 = anythingButSpace
            {
               sb.append("*").append(txt1);
            }
   
            // space & anything
         |  WS AMPER txt1 = anythingButSpace
            {
               sb.append("&").append(txt1);
            }
   
         | (WS)?
           (
                // [space] "text"
                QUOTE txt1 = anythingButQuote QUOTE
                {
                   sb.append(txt1);
                }
                
                // [space] text anything
              | txt1 = anyfile txt2 = anythingButSpace
                {
                   sb.append(txt1).append(txt2);
                }
           )
      )
      {
         str = sb.toString();
      }
   ;

/**
 * Matches a sequence of characters until encountering a closing brace or
 * quote and returns the sequence.
 *
 * @return   The matching sequence of characters as a string.
 */
anythingButQuote returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      str = "";
   }
   :
      (
         options { greedy=true; }
         :
           s1:EQUALS { sb.append(s1.getText()); }
         | s2:STAR   { sb.append(s2.getText()); }
         | s3:AMPER  { sb.append(s3.getText()); }
         | s4:DIGITS { sb.append(s4.getText()); }
         | s5:CODE   { sb.append(s5.getText()); }
         | s6:WS     { sb.append(s6.getText()); }
      )*
      {
         str = sb.toString();
      }
   ;

/**
 * Matches a sequence of characters (including two sequential double quote
 * characters which is an escaping mechanism) until encountering a closing brace
 * or unescaped double quote and returns the sequence.
 *
 * @return   The matching sequence of characters as a string.
 */
anythingButUnescapedQuote returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      str = "";
   }
   :
      (
         options { greedy=true; }
         :
           s1:EQUALS   { sb.append(s1.getText()); }
         | s2:STAR     { sb.append(s2.getText()); }
         | s3:AMPER    { sb.append(s3.getText()); }
         | s4:DIGITS   { sb.append(s4.getText()); }
         | s5:CODE     { sb.append(s5.getText()); }
         | s6:WS       { sb.append(s6.getText()); }
         
           // only one double quote char on output
         | QUOTE QUOTE { sb.append("\""); }
      )*
      {
         str = sb.toString();
      }
   ;
   
/**
 * Matches a sequence of characters that is distinguishable from an argument
 * reference and returns the sequence.
 *
 * @return   The matching sequence of characters as a string.
 */
anyfile returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      str = "";
   }
   :
      (
           s1:EQUALS  { sb.append(s1.getText()); }
         | s2:CODE    { sb.append(s2.getText()); }
         | s3:DIGITS  { sb.append(s3.getText()); }
      )
      {
         str = sb.toString();
      }
   ;

/**
 * Matches a sequence of characters until encountering a closing brace or
 * space and returns the sequence.
 *
 * @return   The matching sequence of characters as a string.
 */
anythingButSpace returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      str = "";
   }
   :
      (
         options { greedy=true; }
         :
           s1:EQUALS  { sb.append(s1.getText()); }
         | s2:STAR    { sb.append(s2.getText()); }
         | s3:AMPER   { sb.append(s3.getText()); }
         | s4:DIGITS  { sb.append(s4.getText()); }
         | s5:CODE    { sb.append(s5.getText()); }
         | s6:QUOTE   { sb.append(s6.getText()); }
      )*
      {
         str = sb.toString();
      }
   ;

/**
 * Matches arguments portion of the include file references and puts the
 * parsed arguments into a container.
 * <p>
 * The first argument determines the parsing strategy. If positional, then
 * all subsequent arguments (even if they have the &amp;name=value form)
 * will be parsed as positional (so the &amp;name= text will be part of the
 * argument). If the first argument is named (prefixed with &amp;name=),
 * then any stray text in between properly formatted named arguments will
 * be parsed as positional arguments and discarded as trash. See the rule
 * {@link #argvalue} for details on how the discard mode works.
 * <p>
 * Core parsing subroutines are {@link #keyarg} and {@link #posarg}.
 *
 * @param    args
 *           Container for arguments.
 */
arguments[Map args]
   {
      int     npos       = 1;
      boolean first      = true;
      boolean positional = false;
      boolean ignore     = false;
   }
   :
      (
         options { generateAmbigWarnings=false; }
         :
         {
            (LA(1) == WS && LA(2) != RBRACE) ||
            (LA(1) != WS && LA(1) != RBRACE)
         }?
         (
            options { generateAmbigWarnings=false; }
            :
            WS
         )?
         (
            options { generateAmbigWarnings=false; }
            :
            
              // &name=value form, only possible to match this on the
              // first pass through the loop (will be bypassed if the
              // next token is not ampersand) or if the first argument
              // previously found was named
              
              { !positional }?
              keyarg[args, npos]
              {
                 // if we are in named mode, then every named argument
                 // found is valid and we need to increment our argument
                 // index
                 npos++;
              }
            
              // positional form, can be matched whether we are 
              // processing positional or named BUT the resulting
              // parsed argument is only saved into the container if
              // we are in positional mode, otherwise the argument is
              // just trash and we discard it (based on the ignore flag)
            | posarg[args, npos, ignore]
              {
                 // track if we should be parsing in positional mode
                 // (only allowed if the first arg is positional and
                 // not named), if so we can never parse the named form
                 // and any argument that has the named format is really
                 // just a positional argument and the resulting text
                 // will include the &name= portion as part of the text
                 if (first)
                    positional = true;
                    
                 // if we are not discarding the arguments then we must
                 // increment our argument index
                 if (!ignore)
                    npos++;
              }
         )
         {
            // no longer the first pass through this loop
            first = false;
            
            // determines if positional arguments (those that are 
            // not prefixed by an &name=) should be honored or
            // discarded (discard if ignore == true); we never
            // discard if this is the first time through or if on
            // the first pass through the loop that first argument
            // was found to be positional
            ignore = !positional;
         }
      )+
      (WS)?
   ;

/**
 * Matches a named argument and puts the parsed argument name and value
 * into a container. The first token encountered must always be an
 * <code>AMPER</code>.  The argument name must follow this and the
 * {@link #anythingButEquals} rule parses that name.  The resulting text has
 * all spaces removed before the name is stored.
 * <p>
 * For a valid value to be found, following the name must be an
 * <code>EQUALS</code> token. Finally, the value is parsed by the rule
 * {@link #argvalue}.  The resulting name and value pair is stored in the
 * given container.
 * <p>
 * If no <code>EQUALS</code> token is found after the name, then this named
 * argument and all other text until the closing <code>RBRACE</code> is
 * dropped.  In such a case, even if there are properly formed named
 * arguments following the malformed one, these are all discarded.
 * <p>
 * If a named argument is matched that is a duplicate of a previously matched
 * named argument, then the previously matched value will be honored and all
 * subsequent instances of same name are discarded. 
 *
 * @param    args
 *           The container where the argument is put.
 * @param    npos
 *           The position of this argument in the argument list.
 */
keyarg[Map args, int npos]
   {
      String  name   = null;
      String  value  = null;
      boolean equals = true; 
   }
   :
      // named argument is &name=value
      // note: this will delete all whitespace from the name (both embedded
      // and trailing!)
      AMPER name = anythingButEquals[true]
      (
         options { generateAmbigWarnings=false; }
         :
           // white spaces are allowed in &a = b even after the EQUAL sign
           {
              (LA(2) == WS && LA(3) != RBRACE) ||
              (LA(2) != WS && LA(2) != RBRACE)
           }?
           EQUALS (WS)? value = argvalue[false]
           
           // malformed named arguments cause all following arguments to be
           // silently discarded
         | (
              options { generateAmbigWarnings=false; }
              :
              ~RBRACE { equals = false; }
           )*
      )
      {
         // check if this was properly formed 
         if (equals)
         {
            // the first instance of this given name is a valid named arg
            if (args.get(name) == null)
            {
               // add our mapping
               args.put(name, value);
            }
            else
            {
               // duplicate names are "dropped" as named arguments but must
               // still be available as a positional argument and as part of
               // the {*} expansion; we create a placeholder here for this
               // case and the argument normalization downstream is smart
               // about handling this edge case
               args.put("{" + npos, value);
            }
         }
      }
   ;

/**
 * Matches a positional argument and puts the parsed argument name and
 * value into a container. The core parsing is handled by {@link #argvalue}.
 * <p>
 * Names for the positional arguments are special combinations made of
 * the opening brace character and the position number. This rule creates
 * a separate namespace for the positional arguments.
 *
 * @param    args
 *           The container where the argument is put.
 * @param    npos
 *           The position of this argument in the argument list.
 * @param    ignore
 *           <code>true</code> to force any matched argument to be discarded
 *           instead of added to the container. This is needed to support
 *           "trash" (that look like positional args) intermixed with named
 *           argument processing.
 */
posarg[Map args, int npos, boolean ignore]
   {
      String name  = "{" + npos;
      String value = null;
   }
   :
      value = argvalue[ignore]
      {
        if (!ignore && value.length() > 0)
           args.put(name, value);
      }
   ;

/**
 * Matches a sequence of characters until encountering a closing brace or
 * '=' and returns the sequence.
 *
 * @param    drop
 *           <code>true</code> to drop whitespace from the returned text.
 *
 * @return   The matching sequence of characters as a string.
 */
anythingButEquals [boolean drop] returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      str = "";
   }
   :
      (
         options { greedy=true; }
         :
           s1:WS        { if (!drop) sb.append(s1.getText()); }
         | s2:STAR      { sb.append(s2.getText()); }
         | s3:AMPER     { sb.append(s3.getText()); }
         | s4:DIGITS    { sb.append(s4.getText()); }
         | s5:CODE      { sb.append(s5.getText()); }
         | s6:QUOTE     { sb.append(s6.getText()); }
      )*
      {
         str = sb.toString();
      }
   ;

/**
 * Matches the argument value portion of either positional or named argument
 * and returns the value. The value of a valid argument extends to (but does
 * not include) the first unquoted whitespace or to the next right brace
 * character.
 * <p>
 * When parsing an argument's value any preceding whitespace is dropped
 * though it is not required to have preceding whitespace. At that point,
 * normal parsing of the value occurs.  In this parsing, a double quote
 * character will cause a special sub-parsing mode to occur where ampersands,
 * equal signs and white space (characters that normally can't be part of
 * an argument value) can be inserted into the argument value. This mode 
 * continues until a closing double quote is encountered OR until the right
 * brace is encountered. This does mean that the closing double quote is
 * optional, but in such a case it will consume everything up to the end
 * of the include specification.
 * <p>
 * If not in the special quoted mode, the normal value parsing will continue
 * until encountering whitespace or a double quote.  Whitespace ends the
 * parsing of a value and the captured text is returned. A double quote  
 * character will drop into the special quoted sub-parsing mode described
 * above. Sequences of quoted "strings" and random characters can be
 * freely intermixed to create the value for an argument. The only issue is
 * that any whitespace that exists outside of the double quoted section will
 * mark the end of the value.
 * <p>
 * It is valid to match nothing in this rule (no whitespace and no value) or
 * to only match whitespace.  In such a case, the parsing will end if an
 * <code>RBRACE</code> is encountered. This is like having a value that is
 * an empty string. 
 * <p>
 * In named parsing mode (if the first argument matched was a named
 * argument in the form &amp;name=value), this method is also used to match  
 * and discard any text in between valid named arguments. To mark this case,
 * the <code>discard</code> will be <code>true</code>. This changes the
 * parsing to process the so called "discard region". This region occurs 
 * after the first unquoted whitespace which marks the end of the last valid
 * named argument. Anything after this point that is not an ampersand or
 * right brace will be discarded.  There is no discard region if positional
 * parsing mode is active (the first argument matched was a positional
 * argument). This means that this discard mode will never be enabled for
 * positional parsing mode.  This means that <code>discard</code> should
 * only be <code>true</code> if this is named parsing mode AND ALSO the
 * parser has entered the discard region.
 * <p>
 * In the discard region (between valid named arguments), double quotes lose
 * their special behavior.  Specifically, in the discard region the double
 * quote character is treated as any other and cannot be used to "hide"
 * ampersands or whitespace as it normally would.  There is no concept of
 * paired sets of double quotes in this part of the parsing.    
 * <p>
 * Once encountering the discard region, whitespace is not required in the 
 * character immediately preceding the next ampersand, but rather as soon as
 * the ampersand is encountered, the discard region is exited and the
 * ampersand is treated as the beginning of the next valid named argument.
 *
 * @param    discard
 *           If <code>true</code>, enter the special discard processing
 *           mode.  This should only be <code>true</code> if the parser is
 *           in both named arguments mode AND a valid named argument value 
 *           has just been parsed.
 *
 * @return   The argument value as a string.
 */
argvalue [boolean discard] returns [String str]
   {
      StringBuilder sb           = new StringBuilder();
      String        txt1         = "";
      boolean       noCloseQuote = false;
      str = "";
   }
   :
      // match nothing (if we read an RBRACE next) or with some valid or
      // invalid (discardable) content
      (
           { discard }?
           (
              options { generateAmbigWarnings=false; }
              :
              txt1 = anythingButSpaceOrAmpersand
           )+
         
         | { !discard }?
           (
              options { generateAmbigWarnings=false; }
              :
              
                // double quotes by themselves leave one quote behind
                QUOTE QUOTE  { sb.append("\""); }
              
                // "text & =text"
                // "text & =text
                // """something"""
                // "some text with ""embedded quotes"""
              | ( QUOTE txt1 = anythingButUnescapedQuote { noCloseQuote = true; } )
                (
                   options { generateAmbigWarnings=false; }
                   :
                   QUOTE { noCloseQuote = false; }
                )?
                {
                   int len = txt1.length(); 
                
                   sb.append(txt1);
                   
                   // TODO: At least 1 case has been found where this is
                   //       deficient. This was an inclref inside an &ELSE block
                   //       whose end looked something like this:
                   //
                   //    &the_arg_name   = "some_text}
                   // &ENDIF
                   //
                   // The some_text parsed fine and we see the RBRACE as the
                   // LA(1). The problem is that in the lexer, the next 2
                   // characters (lexer LA1 and LA2) are \n (correct) and E
                   // (incorrect). The & (ampersand) is missing. Removing the
                   // len == 0 check did not fix it.  Adding parser LA(3) text
                   // to the parserLookahead did not change it either. The
                   // result is that the &ENDIF is not parsed properly because
                   // it is right after the RBRACE and so the text.g condtext
                   // rule just reads to the end of the file until it decides
                   // to die.
                   if (len == 0 && noCloseQuote && LA(1) == RBRACE)
                   {
                      // we only matched a single quote char but in order
                      // to get here we unfortunately loaded LA(2) with
                      // content from the stream which needs to be pushed
                      // back
                      Token la2 = LT(2);
                      parserLookahead = la2.getText();
                   }
                }
               
                // text&=etc
              | txt1 = anythingButSpaceOrQuote  { sb.append(txt1); }
           )+
      )         
      {
         if (sb.length() > 0)
            str = sb.toString();
      }
   ;

/**
 * Matches a sequence of characters until encountering a closing brace,
 * space or double quote character and returns the sequence.
 *
 * @return   The matching sequence of characters as a string.
 */
anythingButSpaceOrQuote returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      str = "";
   }
   :
      (
           s1:EQUALS  { sb.append(s1.getText()); }
         | s2:STAR    { sb.append(s2.getText()); }
         | s3:AMPER   { sb.append(s3.getText()); }
         | s4:DIGITS  { sb.append(s4.getText()); }
         | s5:CODE    { sb.append(s5.getText()); }
      )
      {
         str = sb.toString();
      }
   ;

/**
 * Matches a sequence of characters until encountering a closing brace,
 * ampersand, space or quote and returns the sequence.
 *
 * @return   The matching sequence of characters as a string.
 */
anythingButSpaceOrAmpersand returns [String str]
   {
      StringBuilder sb = new StringBuilder();
      str = "";
   }
   :
      (
           s1:EQUALS  { sb.append(s1.getText()); }
         | s2:STAR    { sb.append(s2.getText()); }
         | s3:DIGITS  { sb.append(s3.getText()); }
         | s4:CODE    { sb.append(s4.getText()); }
         | s5:QUOTE   { sb.append(s5.getText()); }
      )
      {
         str = sb.toString();
      }
   ;

/**
 * Tokenizes the input stream of characters from the portion of Progress
 * source file between braces and returns tokens to the caller according
 * to the needs of preprocessor.
 * <p>
 * There is only a single context for this lexer as no comments are honored
 * inside the braces and strings are treated in a special way: only double
 * quoted strings are recognized, they can be adjacent to text and even
 * unclosed.
 *
 * @see  BracesParser
 */
class BracesLexer extends Lexer;

options
{
   k = 2;                                 // lookahead depth
   charVocabulary = '\u0000'..'\ufffe';   // specifies range 0x0000-0xFFFE
                                          // (0xFFFF is EOF)
}

{
   /** keeps the reference to the shared environment */
   private Environment env = null;
   
   /** Flag to denote if the backslash is honored as an escape sequence. */
   private boolean isUnixEscapes = false;

   /**
    * Constructor. Creates a lexer attached to the input stream
    * taken from the environment. Saves the environment for future needs.
    *
    * @param    env
    *           Shared preprocessor environment.
    */
   public BracesLexer(Environment env)
   {
      this(new CharBuffer(env.getIns()));
      this.env = env;
      this.env.setLsi(this.getInputState());
      this.isUnixEscapes = env.getOpt().isUnixEscapes();
   }
}

/**
 * Matches any amount of whitespace on input.
 * <p>
 * Spaces, tabs and newlines are all matched.  This is a top level lexer rule
 * which means that there is an associated <code>WS</code> token.
 */
WS
   :
      (
         options { greedy=true; }
         :
           ' '                           // space
         | '\t'                          // tab
         | '\n'                          // newline
      )+
      
      // GES: \r can be seen from ClearStream, only if it is escaped, but
      // this lexer will not match it anywhere --> this is a problem
   ;

/**
 * Matches a single equal character.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>EQUALS</code> token.
 */
EQUALS
   :
      '='
   ;

/**
 * Matches a single asterisk character.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>STAR</code> token.
 */
STAR
   :
      '*'
   ;

/**
 * Matches a single ampersand character.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>AMPER</code> token.
 */
AMPER
   :
      '&'
   ;

/**
 * Matches a single quote (not apostrophe) character.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>QUOTE</code> token.
 */
QUOTE
   :
      '"'
   ;

/**
 * Matches any number of successive decimal digits on input.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>DIGITS</code> token.
 */
DIGITS
   :
      (DIGIT)+
   ;

/**
 * Matches a preprocessor input that has no other interpretation. Due to the
 * fact that escaped '}' characters must be ignored here, this rule must
 * include matching for the <code>RBRACE</code> token. Note that the
 * {@link ClearStream} implementation returns the escape character instead
 * of "eating" it when it is parsing braces. The escape character is either
 * tilde or backslash (but backslash is only honored if UNIX escapes are
 * honored) and must be detected here to differentiate between text and
 * the <code>RBRACE</code>.  If an escaped right brace is matched, note that
 * the escape character is dropped unless we are parsing inside a string
 * literal (further up the stack).
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>CODE</code> token.
 */
CODE
   :
           // only match RBRACE if it is the first character seen (and we
           // return with only that character as the text); if the RBRACE
           // character is found but it is not the first character matched
           // in this rule, then we consider this the end of any previous
           // CODE token parsing and return;
           '}'
           {
              $setType(RBRACE);
           }
      
      | (
           options { generateAmbigWarnings=false; }
           :
           
             { env.isInString() }?
             '~' '}'
             
           | { !env.isInString() }?
             '~'! '}'
           
             // turn an escaped tilde into one tilde on output
           |'~'! '~'
           
             // turn an escaped backslash into one backslash on output
           | '~'! '\\'
             
           | '~'
           
           | { env.isInString() && isUnixEscapes }?
             '\\' '}'
           
           | { !env.isInString() && isUnixEscapes }?
             '\\'! '}'
           
             // turn an escaped backslash into one backslash on output
           | { isUnixEscapes }?
             '\\'! '\\'
           
             // turn an escaped tilde into one tilde on output
           | { isUnixEscapes }?
             '\\'! '~'
             
           | '\\'
           
           | ~(
                  ' '
                | '\t'
                | '\r'
                | '\n'
                | '='
                | '*'
                | '&'
                | '"'
                | '0'..'9'
                | '\\'
                | '~'
                | '}'
             )
        )+
   ;
   
/**
 * Matches a single decimal digit on input.
 */
protected
DIGIT
   :
      '0'..'9'
   ;
