/*
** Module   : expression_evaluator.g
** Abstract : tree parser grammar for evaluating Progress 4GL expressions
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20041101   @18718 First version with support for:
**                           - walking an entire tree from the
**                             ProgressParser.expr() function
**                           - for each node, the operator is examined
**                             and each operand is evaluated
**                           - by rationalizing all return types as an
**                             Object, the evaluate() method is made
**                             recursive
**                           - when evaluate() reaches an operand leaf
**                             node, it returns the node's data wrapped 
**                             in a Java non-primitive object and cast
**                             into an Object
**                           - when the full expected list of operand
**                             nodes have all returned, the operator's 
**                             processing occurs and that return data is
**                             wrapped in a Java non-primitive object
**                             and cast into an Object which is returned
**                             to its' caller
**                           - some of the node processing rules are
**                             trivial enough to inline all Java code
**                             necessary to calculate the return data
**                           - other rules require helpers, some of which
**                             are in a separate class and some of which 
**                             are manual additions to this class
**                           - all operators and features of expression
**                             processing for integer, decimal, logical
**                             and character data types in Progress are
**                             supported (date, recid, rowid and handle
**                             are not supported)
** 002 GES 20050426   @20887 Modified string processing to centralize it
**                           in the util package.
** 003 NVS 20050519   @21241 Added SYMBOL alternative to avoid parser's
**                           error message "undefined token" that previou-
**                           sly used to appear for every undefined symbol
**                           tested in the Progress source with DEFINED()
**                           preprocessor built-in function.
** 004 GES 20050526   @21313 Converted StringComparator reference to the
**                           character class.
** 005 GES 20070319   @32498 Rewrite to use the wrapper classes and the
**                           full 4GL semantics to expression processing.
**                           Added the full set of built-in functions used
**                           by the preprocessor. These changes ensure
**                           the results are consistent with how the
**                           4GL would work (the previous implementation
**                           was done before the full base language
**                           semantics had been completed).
** 006 GES 20071107   @35898 Remove schema dictionary loading from the
**                           expression evaluator. This eliminates the
**                           case where preprocessor definitions get
**                           mistaken as field names.
** 007 GES 20081002   @40032 Signature changes.
** 008 GES 20090424   @41955 Import change.
** 009 GES 20090518   @42406 Import change.
** 010 GES 20110712          Added support for the built-in functions KEYWORD
**                           and KEYWORD-ALL.
** 011 GES 20111129          Allow undefined preprocessor symbols to be a
**                           DB_SYMBOL or FILENAME.
*/ 

header
{
/*
** Module   : ExpressionEvaluator.java
**            ExpressionEvaluatorTokenTypes.java
** Abstract : provides tree walking to drive Progress 4GL expression
**            evaluation
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20041116 ADD   @18718 WARNING, THIS IS A GENERATED FILE!!!
**                               DO NOT EDIT THIS FILE. The original source
**                               file is expression_evaluator.g!
*/ 

package com.goldencode.p2j.uast;

import java.io.*;
import java.lang.reflect.*;
import antlr.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.util.*;
}

/**
 * Recursively walks an AST generated from a valid Progress 4GL expression 
 * and evaluates each node to obtain the final result of the expression
 * which is returned to the caller as a <code>BaseDataType</code> (wrapper).
 * <p>
 * The caller generates a tree is by calling the
 * <code>{@link ProgressParser#expr}</code> method.  The 
 * <code>{@link #evaluate}</code> method is called and this tree as passed as
 * a parameter.
 * <p>
 * Starting at the root of this tree, each node is visited recursively (in a
 * depth-first then left to right traversal).  Each node that is visited
 * must be an operator, built-in function, variable (one of the
 * non-parenthesized built-in functions like <code>OPSYS</code>) or a literal
 * (constant).  When a leaf node is encountered, a result is calculated 
 * and returned.  Once all children of a node are evaluated, that node's
 * value is calculated and then returned as the result from that portion of
 * the tree (each child of an operator is an operand and is evaluated before
 * the operator's result can be calculated).  Children of function nodes are
 * the function parameters.
 * <p>
 * By rationalizing all return types as an <code>BaseDataType</code>, the
 * evaluate method is made recursive. This recursive model continues until
 * all subsequent calls to evaluate return. This final result is returned to
 * the original caller as the result of the expression evaluation.
 * <p>
 * The core of the grammar is a single rule called <code>evaluate</code>.
 * <p>
 * Some of the node processing rules are trivial enough to inline all Java
 * code necessary to calculate the return data.  Other rules require helpers,
 * some of which are in the <code>util</code> package (all of the Progress
 * 4GL compatible wrapper processing) and one of which is a manual addition
 * to this class (<code>executeCallback</code>). 
 * <p>
 * All operators and features of expression processing for integer, decimal,
 * logical, date and character data types in Progress are supported. All
 * literal types (including unknown value) are supported.
 * <p>
 * Progress recid, rowid and handle expressions are not supported at
 * this time.
 * <p>
 * The following operators are supported:
 * <p>
 * <pre>
 * + (addition, unary plus, date and string concatination)
 * - (subtraction, unary minus and date)
 * *
 * /
 * modulo
 * ()
 * or
 * and
 * not
 * eq
 * ne
 * gt
 * gte
 * lt
 * lte
 * begins
 * matches
 * </pre>
 * <p>
 * The only operators that are not supported are assignment and the where
 * clause <code>contains</code>.
 * <p>
 * The following built-in functions are supported:
 * <p>
 * <pre>
 * absolute
 * asc
 * date
 * day
 * decimal
 * encode
 * entry
 * etime
 * exp
 * fill
 * index
 * integer
 * keyword
 * keyword-all
 * lc
 * left-trim
 * length
 * log
 * lookup
 * maximum
 * minimum
 * month
 * num-entries
 * opsys
 * propath
 * proversion
 * r-index
 * random
 * replace
 * right-trim
 * round
 * sqrt
 * string
 * substitute
 * substring
 * time
 * today
 * trim
 * truncate
 * weekday
 * year
 * </pre>
 * <p>
 * All data is returned as an <code>BaseDataType</code>, so one must
 * determine what the real return type is before using the returned data.
 * Example:
 * <pre>
 * BaseDataType result = ExpressionEvaluator.evaluateExpr(expr, null, false);
 *   
 * if (result instanceof integer)
 * {
 *    int myResult = ((integer) result).intValue();
 *    do something here...
 * }
 * </pre>
 *
 * @author GES
 */
class ExpressionEvaluator extends TreeParser;

options
{
   k = 1;
   importVocab = ProgressParser; // we must share the same token types as the
                                 // Parser that we are using
}

{
   /** Debug mode enabled. */
   private boolean debug = false;
   
   /**
    * Stores definitions of user-defined functions/variables. May be
    * <code>null</code> if no user-defined features are in use.
    */
   private CallbackResolver callResolver;
   
   /**
    * Constructs an instance allowing the helper classes to be provided by
    * the caller.
    *
    * @param    resolve
    *           Dictionary of <code>Callback</code> definitions for user-
    *           defined functions or variables.
    */
   public ExpressionEvaluator(CallbackResolver resolve)
   {
      this();
      callResolver = resolve;
   }

   /**
    * Returns the status of debug mode.
    *
    * @return    Debug mode enabled/disabled.
    */
   public boolean isDebug()
   {
      return debug;
   }
   
   /**
    * Sets the status of debug mode.
    *
    * @param    dbg
    *           <code>true</code> to enable debug mode, <code>false</code>
    *           to disable.
    */
   public void setDebug(boolean dbg)
   {
      debug = dbg;
   }
         
   /**
    * Provides a command line interface for an end user to drive and/or test
    * this class.
    * <p>
    * Syntax:
    * <pre>
    *   java.exe ExpressionEvaluator &quot;quoted_expression&quot; [debug]
    * </pre>
    * Where:
    * <ul>
    *   <li> &quot;quoted_expression&quot; is a valid Progress 4GL 
    *        expression that is surrounded by double quotes. Note: if strings
    *        must be specified as part of the expression, specify those
    *        strings using the single quote character. 
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args) 
   {
      boolean debug  = false;
      String  syntax = "Syntax: java.exe ExpressionEvaluator " +
                       "<quoted_expression> [debug]"; 
      
      if (args.length < 1)
      {
         System.err.println(syntax);
         System.exit(-1);
      }
      
      if (args.length == 2)
      {
         debug = true;
      }
      
      try
      {
         BaseDataType result = ExpressionEvaluator.evaluateExpr(args[0],
                                                                null,
                                                                debug);
         System.out.println("Result = " + result);         
      }
      
      catch (Exception excpt)
      {
         excpt.printStackTrace();
      }
   }
   
   /**
    * Add the list of built-in functions that are supported in the
    * preprocessor so that these functions can be used in preprocessor 
    * expressions.
    *
    * @param    sym
    *           The parser's symbol resolver.
    * @param    cr
    *           The container managing all function callbacks and variable
    *           references.
    */
   public static void addStandardFunctions(SymbolResolver   sym,
                                           CallbackResolver cr)
   {
      cr.addCallback(sym.lookupKeyword("absolute"),
                     null,
                     "com.goldencode.p2j.util.MathOps",
                     "abs");
      cr.addCallback(sym.lookupKeyword("asc"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "asc");
      cr.addCallback(sym.lookupKeyword("date"),
                     null,
                     "com.goldencode.p2j.util.date",
                     ""); // empty string means this is a constructor 
      cr.addCallback(sym.lookupKeyword("day"),
                     null,
                     "com.goldencode.p2j.util.date",
                     "day");
      cr.addCallback(sym.lookupKeyword("decimal"),
                     null,
                     "com.goldencode.p2j.util.decimal",
                     ""); // empty string means this is a constructor 
      cr.addCallback(sym.lookupKeyword("encode"),
                     null,
                     "com.goldencode.p2j.util.SecurityOps",
                     "encode");
      cr.addCallback(sym.lookupKeyword("entry"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "entry");
      cr.addCallback(sym.lookupKeyword("etime"),
                     null,
                     "com.goldencode.p2j.util.date",
                     "elapsed");
      cr.addCallback(sym.lookupKeyword("exp"),
                     null,
                     "com.goldencode.p2j.util.MathOps",
                     "pow");
      cr.addCallback(sym.lookupKeyword("fill"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "fill");
      cr.addCallback(sym.lookupKeyword("index"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "indexOf");
      cr.addCallback(sym.lookupKeyword("integer"),
                     null,
                     "com.goldencode.p2j.util.integer",
                     ""); // empty string means this is a constructor 
      cr.addCallback(sym.lookupKeyword("keyword"),
                     null,
                     "com.goldencode.p2j.uast.ProgressAst",
                     "keyword"); 
      cr.addCallback(sym.lookupKeyword("keyword-all"),
                     null,
                     "com.goldencode.p2j.uast.ProgressAst",
                     "keywordAll"); 
      cr.addCallback(sym.lookupKeyword("lc"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "toLowerCase");
      cr.addCallback(sym.lookupKeyword("left-trim"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "leftTrim");
      cr.addCallback(sym.lookupKeyword("length"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "length");
      cr.addCallback(sym.lookupKeyword("log"),
                     null,
                     "com.goldencode.p2j.util.MathOps",
                     "log");
      cr.addCallback(sym.lookupKeyword("lookup"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "lookup");
      cr.addCallback(sym.lookupKeyword("maximum"),
                     null,
                     "com.goldencode.p2j.util.CompareOps",
                     "maximum");
      cr.addCallback(sym.lookupKeyword("minimum"),
                     null,
                     "com.goldencode.p2j.util.CompareOps",
                     "minimum");
      cr.addCallback(sym.lookupKeyword("month"),
                     null,
                     "com.goldencode.p2j.util.date",
                     "month");
      cr.addCallback(sym.lookupKeyword("num-entries"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "numEntries");
      cr.addCallback(sym.lookupKeyword("opsys"),
                     null,
                     "com.goldencode.p2j.util.EnvironmentOps",
                     "getOperatingSystem");
      cr.addCallback(sym.lookupKeyword("propath"),
                     null,
                     "com.goldencode.p2j.util.EnvironmentOps",
                     "getSearchPath");
      cr.addCallback(sym.lookupKeyword("proversion"),
                     null,
                     "com.goldencode.p2j.util.EnvironmentOps",
                     "getVersion");
      cr.addCallback(sym.lookupKeyword("r-index"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "lastIndexOf");
      cr.addCallback(sym.lookupKeyword("random"),
                     null,
                     "com.goldencode.p2j.util.MathOps",
                     "random");
      cr.addCallback(sym.lookupKeyword("replace"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "replaceAll");
      cr.addCallback(sym.lookupKeyword("right-trim"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "rightTrim");
      cr.addCallback(sym.lookupKeyword("round"),
                     null,
                     "com.goldencode.p2j.util.MathOps",
                     "round");
      cr.addCallback(sym.lookupKeyword("sqrt"),
                     null,
                     "com.goldencode.p2j.util.MathOps",
                     "sqrt");
      cr.addCallback(sym.lookupKeyword("string"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "valueOf");
      cr.addCallback(sym.lookupKeyword("substitute"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "substitute");
      cr.addCallback(sym.lookupKeyword("substring"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "substring");
      cr.addCallback(sym.lookupKeyword("time"),
                     null,
                     "com.goldencode.p2j.util.date",
                     "secondsSinceMidnight");
      cr.addCallback(sym.lookupKeyword("today"),
                     null,
                     "com.goldencode.p2j.util.date",
                     ""); // empty string means this is a constructor
      cr.addCallback(sym.lookupKeyword("truncate"),
                     null,
                     "com.goldencode.p2j.util.MathOps",
                     "truncate");
      cr.addCallback(sym.lookupKeyword("trim"),
                     null,
                     "com.goldencode.p2j.util.character",
                     "trim");
      cr.addCallback(sym.lookupKeyword("weekday"),
                     null,
                     "com.goldencode.p2j.util.date",
                     "weekday");
      cr.addCallback(sym.lookupKeyword("year"),
                     null,
                     "com.goldencode.p2j.util.date",
                     "year");
   }
   
   /**
    * External entry point to evaluates a Progress 4GL expression as 
    * expressed in a string of source code.
    * <p>
    * Optionally user-defined functions and variables can be provided via
    * a <code>CallbackResolver</code> object.
    * <p>
    * All data is returned as a <code>BaseDataType</code>, so one must
    * determine what the real return type is before using the returned data.
    * Example:
    * <pre>
    * BaseDataType result = ExpressionEvaluator.evaluateExpr(expr, null, false);
    *   
    * if (result instanceof integer)
    * {
    *    int myResult = ((integer) result).intValue();
    *    do something here...
    * }
    * </pre>
    *
    * @param    exp
    *           Progress 4GL expression to evaluate.
    * @param    cr
    *           Instance of <code>CallbackResolver</code> configured with 
    *           <code>Callback</code> objects representing the functions 
    *           and variables to resolve.  If set to <code>null</code> then 
    *           no user-defined callbacks will be defined.
    * @param    debug
    *           If true, the expression evaluator will output debugging
    *           information to <code>stdout</code>.  Otherwise no output
    *           is generated.
    *
    * @return   The result of the expression validation or <code>null</code>
    *           on error.
    */
   public static BaseDataType evaluateExpr(String           exp,
                                           CallbackResolver cr,
                                           boolean          debug)
   {                                 
      BaseDataType result = null;
      
      try
      {
         // initial setup (warning, schema dictionary support is disabled
         // but basic lvalue processing in the parser should still be safe
         // since the symbol resolver handles this case)
         SymbolResolver  sym    = new SymbolResolver();
         StringReader    expr   = new StringReader("(" + exp + ")");
         ProgressLexer   lexer  = new ProgressLexer(expr, sym);
         ProgressParser  parser = new ProgressParser(lexer, sym);
         parser.setEvaluatingExpression(true);

         // if one is not provided, we must create one to hold our built-in
         // functions
         if (cr == null)
            cr = new CallbackResolver();
         
         // populate the CR with the standard built-in functions
         // (this MUST be done after the lexer is initialized above otherwise
         // any conflicts will be blown away)
         addStandardFunctions(sym, cr);
         
         // populate SymbolResolver with all Keywords in the CallbackResolver
         cr.populateSymbolResolver(sym);
         
         // parse it and create a tree
         parser.expr();
         CommonAST parseTree = (CommonAST) parser.getAST();
         
         // evaluate the tree
         ExpressionEvaluator expression = new ExpressionEvaluator(cr);
         expression.setDebug(debug);
         
         result = expression.evaluate(parseTree);         
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
      
      return result;
   }
   
   /**      
    * This is the worker function that indirectly calls (through reflection)
    * the method that matches the text in the <code>AST</code>.  This
    * function manually processes a subtree starting at the function's
    * node.  All children are manually walked and the each child is
    * evaluated to obtain the result.  As the depth of the subtree for
    * each child is arbitrary (each parameter can be a valid Progress 4GL
    * expression of arbitrary complexity), the <code>{@link #evaluate}</code> 
    * function is used to recursively process these expressions until the
    * result is returned as an <code>Object</code>.  This is assigned to
    * the function call's array of arguments.  In addition, each of these
    * arguments is queried for its' class and this class is stored in an
    * array that represents the signature of the method.  Once all arguments
    * and the signature are collected, the <code>CallbackResolver</code> is
    * used to invoke the callback that matches the function's text. 
    * <p>
    * On the receiving end, each parameter is unpacked into its' primitive 
    * form if available.  The method executes and returns as it normally
    * would.  Reflection handles returning the result.  Note that the result
    * must be some form of <code>BaseDataType</code>.
    * <p>
    * Note that there are a large number of potential failures in the
    * reflection lookup and execution process.  This method passes these
    * exceptions through.
    * <p>
    * At this time the <code>returnType</code> is not used but it may be
    * used by the <code>lookupCallback</code> in the future. This is here
    * since the Java reflection interfaces are expected to change to require
    * this.
    *
    * @see      #evaluate(AST)
    * @see      CallbackResolver#lookupCallback
    *
    * @param    function 
    *           Node representing the function to be executed.
    * @param    returnType
    *           Return data type represented as a class.
    *
    * @return   Return data from the method call.
    */
   private BaseDataType executeCallback(AST function, Class returnType)
   throws UnsupportedOperationException,
          RecognitionException,
          IllegalAccessException,
          ClassNotFoundException,
          NoSuchMethodException,                              
          InvocationTargetException,
          InstantiationException,
          NullPointerException                                   
   { 
      return executeCallback(function, returnType, false, 0);
   }
   
   /**      
    * This is the worker function that indirectly calls (through reflection)
    * the method that matches the text in the <code>AST</code>.  This
    * function manually processes a subtree starting at the function's
    * node.  All children are manually walked and the each child is
    * evaluated to obtain the result.  As the depth of the subtree for
    * each child is arbitrary (each parameter can be a valid Progress 4GL
    * expression of arbitrary complexity), the <code>{@link #evaluate}</code> 
    * function is used to recursively process these expressions until the
    * result is returned as an <code>Object</code>.  This is assigned to
    * the function call's array of arguments.  In addition, each of these
    * arguments is queried for its' class and this class is stored in an
    * array that represents the signature of the method.  Once all arguments
    * and the signature are collected, the <code>CallbackResolver</code> is
    * used to invoke the callback that matches the function's text. 
    * <p>
    * On the receiving end, each parameter is unpacked into its' primitive 
    * form if available.  The method executes and returns as it normally
    * would.  Reflection handles returning the result.  Note that the result
    * must be some form of <code>BaseDataType</code>.
    * <p>
    * Note that there are a large number of potential failures in the
    * reflection lookup and execution process.  This method passes these
    * exceptions through.
    * <p>
    * At this time the <code>returnType</code> is not used but it may be
    * used by the <code>lookupCallback</code> in the future. This is here
    * since the Java reflection interfaces are expected to change to require
    * this.
    *
    * @see      #evaluate(AST)
    * @see      CallbackResolver#lookupCallback
    *
    * @param    function 
    *           Node representing the function to be executed.
    * @param    returnType
    *           Return data type represented as a class.
    * @param    varargs
    *           <code>true</code> to shift the parameters into an array
    *           form that can be used with method signatures that handle
    *           variable arguments by passing a last parameter as an array
    *           of objects.
    * @param    argIdx
    *           The parameter number (0-based, counting from the "left") with
    *           which to start the variable portion of the arguments (only
    *           used if <code>varargs</code> is <code>true</code>).
    *
    * @return   Return data from the method call.
    */
   private BaseDataType executeCallback(AST     function,
                                        Class   returnType,
                                        boolean varargs,
                                        int     argIdx)
   throws UnsupportedOperationException,
          RecognitionException,
          IllegalAccessException,
          ClassNotFoundException,
          NoSuchMethodException,                              
          InvocationTargetException,
          InstantiationException,
          NullPointerException                                   
   { 
      int      num       = function.getNumberOfChildren();
      Object[] args      = new Object[num];
      Class[]  signature = new Class[num];
      AST      child     = function.getFirstChild(); 
      boolean  undef     = false;

      // build up the function call parameters by evaluating each child
      // in order
      for (int i = 0; i < num; i++)
      {
         args[i] = evaluate(child);
         
         if (args[i] == null)
            undef = true;
         else
            signature[i] = args[i].getClass();
            
         child = child.getNextSibling(); 
      }
     
      if (undef)
      {
         signature = new Class[0];
         args      = new Object[0];
      }

      if (varargs)
      {
         int vsz = args.length - argIdx;
         
         Object[]       bcopy = new Object[argIdx + 1];
         Class[]        scopy = new Class[argIdx + 1];
         BaseDataType[] vcopy = new BaseDataType[vsz];
         
         for (int j = 0; j < args.length; j++)
         {
            if (j < argIdx)
            {
               // the base copy and signature copy will always have the same
               // 0-based indexes as args and signature so j is already
               // correct
               bcopy[j] = args[j];
               scopy[j] = signature[j];
            }
            else
            {
               // the varargs copy may have an offset index position (if
               // argIdx > 0)
               vcopy[j - argIdx] = (BaseDataType) args[j];
            }
         }
      
         // the args have been split (if needed) and now we need to ensure
         // that the variable portion is the last element (if argIdx == 0
         // then this is the only element)
         bcopy[argIdx] = vcopy;
         
         // switch our arguments to the modified version
         args = bcopy;
         
         // ensure the last (or only if argIdx == 0) element of the signature
         // represents the var args (an array of BDT values)
         scopy[argIdx] = (new BaseDataType[0]).getClass();
         
         // switch our signature to the modified version
         signature = scopy;
      }
      
      Object res = callResolver.invokeCallback(function.getText(),
                                               returnType,
                                               signature,
                                               args);
      
      if (!(res instanceof BaseDataType))
      {
         // try some simple conversions
         if (res instanceof Integer)
         {
            res = new integer((Integer) res);
         }
         else if (res instanceof Double)
         {
            res = new decimal((Double) res);
         }
         else if (res instanceof Boolean)
         {
            res = new logical((Boolean) res);
         }
         else if (res instanceof String)
         {
            res = new character((String) res);
         }
         else
         {
            throw new ClassCastException("Can't convert " +
                                         res.getClass().getName() +
                                         "to BaseDataType!");
         }
      }
      
      return (BaseDataType) res;
   }
}

/**
 * Core method that does the following:
 * <ul>
 *   <li> ensures that all alternatives recursively call this same method
 *        for all child nodes
 *   <li> forces the return type to <code>BaseDataType</code> and where
 *        necessary wraps and/or casts the data into the proper non-primitive
 *        objects for a proper return
 *   <li> specifies all possible alternatives and for each one, the proper
 *        processing is either executed inline or is accessed in another
 *        method or class
 * </ul>
 *
 * @param    _t
 *           The tree node to be recursively evaluated.
 *
 * @return   The result of the evaluation of the subtree rooted at the node
 *           passed as a parameter.  This result is wrapped (if necessary) in
 *           a <code>BaseDataType</code> before returning.
 */
evaluate returns [BaseDataType r] 
throws UnsupportedOperationException,
       IllegalAccessException,
       ClassNotFoundException,
       NoSuchMethodException,                              
       InvocationTargetException,
       InstantiationException,
       NullPointerException
   {
      int num = _t.getNumberOfChildren();
      
      // operator processing needs its operands, do this in 1 place
      final AST a = num > 0 ? _t.getFirstChild() : null;
      final AST b = num > 1 ? a.getNextSibling() : null;
         
      BaseDataType op1     = null;
      BaseDataType op2     = null;
      boolean      varargs = false;
      int          ot      = 0;
      int          aindx   = 0;
      
      r = null;
      
      if (isDebug() && _t != null)
         System.out.println(_tokenNames[_t.getType()]);
   }

   :
      // operators
        KW_OR
        {
           LogicalExpression le = new LogicalExpression()
           {
              public logical execute()
              {
                 logical l = null;
                 
                 try
                 {
                    l = (logical) evaluate(b); 
                 }
                 
                 catch (Exception e)
                 {
                    e.printStackTrace(System.err);
                 }
                 
                 return l; 
              }
           };
           r = logical.or((logical) evaluate(a), le);
        }
      | KW_AND
        {
           LogicalExpression le = new LogicalExpression()
           {
              public logical execute()
              {
                 logical l = null;
                 
                 try
                 {
                    l = (logical) evaluate(b); 
                 }
                 
                 catch (Exception e)
                 {
                    e.printStackTrace(System.err);
                 }
                 
                 return l; 
              }
           };
           r = logical.and((logical) evaluate(a), le);
        }
      | KW_NOT  { r = logical.not((logical) evaluate(a)); }
      | EQUALS  { r = CompareOps.isEqual(evaluate(a), evaluate(b)); }
      | NOT_EQ  { r = CompareOps.isNotEqual(evaluate(a), evaluate(b)); }
      | GT      { r = CompareOps.isGreaterThan(evaluate(a), evaluate(b)); }
      | LT      { r = CompareOps.isLessThan(evaluate(a), evaluate(b)); }
      | GTE
        {
           r = CompareOps.isGreaterThanOrEqual(evaluate(a), evaluate(b));
        }
      | LTE
        {
           r = CompareOps.isLessThanOrEqual(evaluate(a), evaluate(b));
        }
      | KW_BEGINS
        {
           r = character.begins((character) evaluate(a),
                                (character) evaluate(b)); 
        }
      | KW_MATCHES 
        {
           r = character.matches((character) evaluate(a),
                                 (character) evaluate(b)); 
        }
      | PLUS
        {
           op1 = evaluate(a);
           op2 = evaluate(b);
           
           if (op1 instanceof NumberType)
           {
              r = MathOps.plus((NumberType) op1, (NumberType) op2);
           }
           else if (op1 instanceof character)
           {
              r = character.concat((character) op1, (character) op2);
           }
           else
           {
              // assume this is a date + number
              if (op2 instanceof integer)
              {
                 r = date.plusDays((date) op1, (integer) op2);
              }
              else
              {
                 r = date.plusDays((date) op1, (decimal) op2);
              }
           }
        }
      | MINUS
        {
           op1 = evaluate(a);
           op2 = evaluate(b);
           
           if (op1 instanceof NumberType)
           {
              r = MathOps.minus((NumberType) op1, (NumberType) op2);
           }
           else
           {
              // assume this is a date - number
              if (op2 instanceof integer)
              {
                 r = date.minusDays((date) op1, (integer) op2);
              }
              else
              {
                 r = date.minusDays((date) op1, (decimal) op2);
              }
           }
        }
      | MULTIPLY
        {
           r = MathOps.multiply((NumberType) evaluate(a),
                                (NumberType) evaluate(b));
        }
      | DIVIDE
        {
           r = MathOps.divide((NumberType) evaluate(a),
                              (NumberType) evaluate(b));
        }
      | KW_MOD
        {
           r = MathOps.modulo((NumberType) evaluate(a),
                              (NumberType) evaluate(b));
        }
      | UN_MINUS      
        {
           op1 = evaluate(a);
           
           if (op1 instanceof integer)
           {
              r = MathOps.negate((integer) op1);
           }
           else
           {
              r = MathOps.negate((decimal) op1);
           }
        }
      | UN_PLUS         { r = evaluate(a); }  // unary plus is meaningless!
      | LPARENS         { r = evaluate(a); }
      
      // literals
      | UNKNOWN_VAL     { r = new unknown(); }
      | n:NUM_LITERAL   { r = new integer(n.getText()); }
      | d:DEC_LITERAL   { r = new decimal(d.getText()); }
      | BOOL_TRUE       { r = new logical(true); }
      | BOOL_FALSE      { r = new logical(false); }
      | da:DATE_LITERAL { r = new date(da.getText()); }
      | s:STRING      
        {
           String str = character.progressToJavaString(s.getText());                                                                
           r = new character(StringHelper.processEscapes(str));
        }
      
      // functions and variables
      | fp:FUNC_POLY
        {
           aindx   = 0;
           varargs = false;
           Aast poly = (Aast) #fp;
        
           if (poly.isAnnotation("oldtype"))
           {
              ot = ((Long) poly.getAnnotation("oldtype")).intValue();
              
              if (ot == KW_MIN || ot == KW_MAX)
              {
                 varargs = true;
              }
           }
        
           r = executeCallback(fp, null, varargs, aindx);
        }
      | fi:FUNC_INT     { r = executeCallback(fi, integer.class); }
      | vi:VAR_INT      { r = executeCallback(vi, integer.class); }
      | fd:FUNC_DEC     { r = executeCallback(fd, decimal.class); }
      | vd:VAR_DEC      { r = executeCallback(vd, decimal.class); }
      | fb:FUNC_LOGICAL { r = executeCallback(fb, logical.class); }
      | vb:VAR_LOGICAL  { r = executeCallback(vb, logical.class); }
      | fda:FUNC_DATE   { r = executeCallback(fda, date.class); }
      | vda:VAR_DATE    { r = executeCallback(vda, date.class); }
      | fc:FUNC_CHAR
        {
           aindx   = 0;
           varargs = false;
           Aast fchar = (Aast) #fc;
        
           if (fchar.isAnnotation("oldtype"))
           {
              ot = ((Long) fchar.getAnnotation("oldtype")).intValue();
              
              if (ot == KW_SUBSTIT)
              {
                 aindx   = 1;
                 varargs = true;
              }
           }
        
           r = executeCallback(fc, character.class, varargs, aindx);
        }
      | vc:VAR_CHAR     { r = executeCallback(vc, character.class); }
      
      // these alternatives are for undefined preprocessor symbols
      // they suppress the unnecessary parser error message
      | SYMBOL
      | DB_SYMBOL
      | FILENAME
   ;
